function nextlevel = emmcluster(varargin)
% EMMCLUSTER implements the EM algorithm for learning mixture hierarchies
% presented by Vasconcelos and Lippman in 
%
% Vasconcelos, N. and Lippman, A., "Learning Mixture Hierarchies", In
% Proceddings of the International Conference on Neural Information
% Processing Systems (NIPS'98), Denver, Colorado, USA, 1998.
%  
% NEXTLEVEL = EMMCLUSTER(BASELAYER,COMPONENTS) computes the gmdistribution
% object NEXTLEVEL with COMPONENTS mixture components based
% on the model(s) given in BASELAYER. BAYESLAYER can either be a
% gmdistribution object, or a cell of gmdistribution objects. 
%
% NEXTLEVEL = EMMCLUSTER(BASELAYER,COMPONENTS,N,NITER) is the same as
% EMMCLUSTER(BASELAYER,COMPONENTS) with the additional parameters N and NITER which
% control the number of virtual samples (see paper) and the number of iterations 
% for the EM algorithm. (Default: N=10, NITER=10);
%
% Example (group a mixture with 16 components into a mixture of 2
% components).
%
%   >> gentest % sample script to generate baselayer model
%   >> nextlayer = emmcluster(gm,2,10)
%
% Author: Roland Kwitt, 2009  

    % check input arguments
    [baselayer,components,dimensionality,N,nIter] = ParseInput(varargin{:});
    % initialize model on next higher level
    nextlevel = randinit(baselayer,components);
    N
    for j=1:nIter
        % compute the new weights, mu and Sigma
        [weights,mu,Sigma] = train(baselayer,nextlevel,dimensionality,N);
        % generate gmdistribution object with new parameters
        nextlevel = gmdistribution(mu,Sigma,weights);
    end
end

% computes the new estimates of the mixture component weightings (weight),
% new mean vectors (mu) and the new covariance matrices (Sigma)
function [weights,mu,Sigma] = train(baselayer,nextlevel,dimensionality,N)
    % #components of baselayer
    I = baselayer.NComponents;
    % #components of nextlayer
    J = nextlevel.NComponents;
    K = J;
    fprintf('Computing common terms ... ');
    M = baselayer.PComponents*N;
    tic;
    for i=1:I
        % implement denumerator of Eq. (10)
        for k=1:K 
            d_partA(i,k) = mvnpdf(baselayer.mu(i,:), ...
                nextlevel.mu(k,:), ...
                diag(nextlevel.Sigma(:,:,k)));
            d_partB(i,k) = exp(-1/2 * trace(inv(diag(nextlevel.Sigma(:,:,k)))*diag(baselayer.Sigma(:,:,i))));
            d_partC(i,k) =  nextlevel.PComponents(k) * (d_partB(i,k) * d_partA(i,k))^M(i);
        end
    end
    denum = sum(d_partC,2);
    time = toc;
    fprintf('done in %0.2f seconds\n', time);

    % compute h_ij all i=1...I and j=1....J, Eq. (10)
    fprintf('Computing h(i,j) ... ');
    tic;
    for i=1:I
        n_Sigmai = diag(baselayer.Sigma(:,:,i));
        n_mui = baselayer.mu(i,:);
        for j=1:J
            n_partA(i,j) = mvnpdf(n_mui,nextlevel.mu(j,:),diag(nextlevel.Sigma(:,:,j)));
            n_partB(i,j) = exp(-1/2 * trace(inv(diag(nextlevel.Sigma(:,:,j)))*n_Sigmai));
            n_partC(i,j) = nextlevel.PComponents(j) * (n_partB(i,j)*n_partA(i,j))^M(i);
            h(i,j) = n_partC(i,j)./denum(i);
        end
    end
    time = toc;
    fprintf('done in %0.2f seconds\n', time);
    
    % determine w(i,j) used in the update step of the parameters
    fprintf('Computing update steps ...');
    tic;
    for i=1:I
        for j=1:J
            w(i,j) = (h(i,j) * baselayer.PComponents(i))/(sum(h(:,j).*baselayer.PComponents(:)));
        end
    end
    SigmaM = zeros(dimensionality,dimensionality,nextlevel.NComponents);
    for j=1:J
        % component weights, Eq. (12)
        weights(j) = sum(h(:,j))./I; 
        % component mean vectors, Eq. (13)
        for i=1:I
            tmp_mu(i,:) = w(i,j) * baselayer.mu(i,:);
            % new covariance matrices, Eq. (14)
            SigmaM(:,:,j) = double(SigmaM(:,:,j)) + ...
                w(i,j) * (diag(baselayer.Sigma(:,:,i)) + ...
                (baselayer.mu(i,:)-nextlevel.mu(j,:))'*(baselayer.mu(i,:)-nextlevel.mu(j,:)));
        end
        Sigma(1,:,j) = diag(SigmaM(:,:,j));
        mu(j,:) = sum(tmp_mu);
    end
    time = toc;
    fprintf('done in %0.2f seconds\n', time);

end

% random init of nextlayer
function Ml = randinit(baselayer,components)
    % draw random sample from baselayer
    X = random(baselayer,1000);
    % initialize mean vectors
    mu = X(randsample(1000,components),:);
    % uniform weights
    weight = ones(1,components)/components ;
    % variance of random sample initializes the covariances
    Sigma = repmat(var(X),[1,1,components]);
    Ml = gmdistribution(mu,Sigma,weight);
end

% create a new gmdistribution object with COMPONENTS GMMs 
% from the gmdistribution ojbects in MODELS
function Ml = assemble(models,dimensionality,components)
    weights = []; mu = []; Sigma = []; 
    nModels = length(models);
    % concatenate models
    for i=1:nModels
        weights = [weights models{i}.PComponents];
        mu = [mu; models{i}.mu];
        Sigma = [Sigma; models{i}.Sigma(:)];
    end
    % ensure positive-definiteness
    p = find (Sigma <= eps*max(Sigma));
    Sigma(p) = eps*max(Sigma);
    % normalize weights
    weights = weights./nModels;
    Sigma = reshape(Sigma,1,dimensionality,nModels*components);
    Ml = gmdistribution(mu,Sigma,weights);
end

function [baselayer,components,dimensionality,N,nIter] = ParseInput(varargin)
    iptchecknargin(1,4,nargin,mfilename);
    % check of first argument is either of type gmdistribution or cell
    baselayer = varargin{1};
    iptcheckinput(baselayer,{'gmdistribution','cell'}, ...
        {'nonempty'},mfilename,'baselayer',1);
    % check if number of new components is nonzero
    components = varargin{2};
    iptcheckinput(components,{'double'},{'nonzero'},mfilename,'components',2);
    % in case the user provides three input arguments the third arg. is N
    if (nargin > 2)
        N = varargin{3};
        iptcheckinput(N,{'double'},{'nonzero'},mfilename,'N',3);
        if (nargin == 4)
            nIter = varargin{4};
            iptcheckinput(nIter,{'double'},{'nonzero'},mfilename,'nIter',4);
        else
            nIter = 10;
        end
    else
        nIter = 10;
        N = 10;
    end
    
    % in case we get a cell of gmdistribution objects, we have to
    % create one new baselayer object composed of all elements of the
    % cell and normalized the weights
    if (strcmp(class(baselayer),'cell'))
        elementcomponents = baselayer{1}.NComponents;
        elementdimensions = baselayer{2}.NDimensions;
        dimensionality = elementdimensions;
        baselayer = assemble(baselayer, ...
            elementdimensions, ...
            elementcomponents);
    else
        dimensionality = baselayer.NDimensions;
    end
end




