function model = emmfe(data,varargin)
% EMMFE Fitting of Embedded Multiresolution Mixture Models (EMM) to DCT
% coefficients using EM algorithm.
%   MODELS = EMMFE(DATA,'PARAM1',val1,'PARAM2', val2, ...) fits EMMs to the
%   images provides by DATA. This code implements the paper:
%   
%   Vasconcelos, N., "A Probabilistic Architecture for
%   Content--Based Image Retrieval". In Proceedings of the IEEE Conference
%   on Computer Vision and Pattern Recognition (CVPR'00), pages 216-221,
%   2000.
%
%   Parameters are:
%
%       DATA                Is a cell array with elements 'image'
%                           'dim', 'idx', 'filename'. 'image' hold the
%                           color or grayscale image. 'dim' specifies the
%                           dimension of the image (in case of color images
%                           'dim' is 3, otherwise 1). 'idx' is not used
%                           here; in case of image classification
%                           it can be used to specify the class the
%                           image belongs to. Finally, 'filename' holds the
%                           image filename.
%         
%       'components'        number of mixture components of the
%                           multivariate Gaussian mixture models which are the
%                           basis of the EMM approach
%
%       'coefficients'      specifies the number of DCT coefficients that
%                           are extracted in each 8x8 block. Coefficients
%                           are extract in zigzag scan order. In case of 
%                           color images the first 1:N coefficients of each 
%                           channel are extracted, and interleaved in the
%                           following way: YBRYBR...; the total number of
%                           coefficients is 3N; in case of grayscale images
%                           the total number is N (default: 1:16)
%
%       'grayscale'         true, if you want to convert the color images in
%                           DATA to grayscale images (only the luminance 
%                           component of the LUV model is kept. In case 
%                           'grayscale' is set to false, the images are converted 
%                           to the YCbCr color model (default: false)
%
%       'lbg'               true, if you want to initialize the EM algorithm
%                           using the Linde-Buzo-Gray algorithm (aka
%                           generalized Lloyd) according to the codeword
%                           splitting procedure of Gray (see Gray, R.
%                           "Vector Quantization", In: IEEE Signal
%                           Processing Magazine, ...; Otherwise, the
%                           default initialization method for the
%                           gmdistribution.fit routine is used (default:
%                           false)
%
%       'preprocesss'       true if you want to use preprocessing of the
%                           input images, i.e. subtraction of the mean and
%                           division by the std. deviation, as proposed by
%                           Do & Vetterli. (default: false)
%
%       'shift'             specifies the number of pixel positions the
%                           sliding 8x8 window is shifted in both
%                           horizontal and vertical direction (default: 2)
%
%       'debug'             true, if you want to obtain debug messages
%                           during computation (default: false)
%
%   Output:
%
%       MODELS              a cell array of gmdistribution objects for
%                           each of the images in DATA
%   Example:
%
%       I = imread('test.bmp');
%       data{1}.image = I; 
%       data{1}.dim = 3; 
%       data{1}.filename = 'test.bmp';
%       model = emmfe(data,'debug',true,'grayscale',true,'lbg',true);
%
%   Author: Roland Kwitt, rkwitt@gmx.at, 2009

    p = inputParser;
    p.addRequired('data',@iscell);
    p.addParamValue('components',8,@(x)x>=1 && x<= 16);
    p.addParamValue('coefficients', 1:16, @isvector);
    p.addParamValue('preprocess',false,@islogical);
    p.addParamValue('grayscale',false,@islogical);
    p.addParamValue('lbg',false,@islogical); % do Linde-Buzo-Gray initialization with codeword splitting
    p.addParamValue('shift',2, @(x)x>=1 && x <= 7);
    p.addParamValue('debug',false,@islogical);
    p.parse(data,varargin{:});
    components = p.Results.components;
    shift = p.Results.shift;
    grayscale = p.Results.grayscale;
    lbg = p.Results.lbg;
    coef = p.Results.coefficients;
    preprocess = p.Results.preprocess;
    debug = p.Results.debug;
    progname = 'emmfe';
    model = {};
    % iterate over all images
    for i=1:length(data)
        if (debug)
            fprintf('%s: processing image %d\n', progname, i);
            tic;
        end
        im = data{i}.image;
        % in case the image is a color image and we want to do
        % preprocessing, we preprocess each channel separately; the images
        % are first converted to the YBR color model
        
        if (size(im,3) > 1) % image is color            
            if (grayscale) % do we want to convert to grayscale
                if (debug)
                    fprintf('%s: converting to grayscale\n',progname);
                end
                im = double(rgb2gray(im)); % convert from RGB to grayscale
                if (preprocess) % do preprocessing
                    if (debug)
                        fprintf('%s: subtracting mean and dividing by std. deviation\n', progname);
                    end
                    im = (im - mean2(im))./std2(im);
                end
            else % leave color image as it is
                if (debug)
                    fprintf('%s: converting to YCbCr\n',progname);
                end
                im = double(rgb2ycbcr(im)); % convert to YBR color model
                if (preprocess) % preprocess each color channel 
                    if (debug)
                        fprintf('%s: subtracting mean and dividing by std. deviation\n', progname);    
                    end
                    for c=1:size(im,3)
                        channel = im(:,:,c);
                        channel = (channel - mean2(channel))./std2(channel);
                        im(:,:,c) = channel;
                    end
                end
            end
        else % image is grayscale image
            if (debug)
                fprintf('%s: image is grayscale\n',progname);
            end
            im = double(im);
            if (preprocess) % preprocess the luminance component
                if (debug)
                    fprintf('%s: subtracting mean and dividing by std. deviation\n', progname);
                end
                im = (im - mean2(im))./std2(im);
            end
        end
        if (debug)
            options = statset('Display','final','MaxIter',500);
        else
            options = statset('MaxIter',500);
        end
        
        % extract DCT coefficients from sliding windows incremented by
        % 'shift' pixel in both horizontal and vertical direction;
        
        block = dct2block(im,8,coef,shift); 
        nLen = size(block,1);
        nDim = size(block,2);
        if (debug)
            fprintf('%s: fitting data block (%d,%d)\n', progname, nLen,nDim);
        end
        if (lbg) % do LBG initialization
            if (debug)
                fprintf('%s: using linde-buzo-gray with codeword splitting\n',progname);
            end
            [mv,assignment] = vqlbg(block',components);
            for a=1:components
                modelI.mu(a,:) = mv(:,a)';
                modelI.Sigma(1,:,a) = var(block(assignment == a,:));
                modelI.PComponents(a) = length(find(assignment == a))/length(assignment);
            end
        end
        
        % in case of LBG initialization it makes no sense to start the EM
        % algorithm several times since the init. procedure is
        % deterministic; in case we use the standard (random)
        % initialization of the gmdistribution.fit routine it makes sense
        % to start the algorithm several times to reach convergence since
        % it depends on the initialization.
        
        if (lbg)
            obj = gmdistribution.fit(block, components, ...
                'CovType','diagonal', ...
                'Options', options, ...
                'Regularize', eps, ...
                'Start',modelI);
        else
            converged = 0; maxTries = 1;
            while (converged ~= 1 && maxTries < 20)
                if (debug)
                    fprintf('%s: %d-th trial\n', progname, maxTries);
                end
                try
                    obj = gmdistribution.fit(block, components, ...
                        'CovType', 'diagonal', ...
                        'Regularize',eps, ...
                        'Options', options);
                    converged = 1;
                catch ME
                    if (debug)
                        fprintf('%s: no success\n', progname);
                    end
                    maxTries = maxTries + 1;
                    obj = [];
                    converged = 0;
                end
            end
        end
        for c=1:components
            if (length(unique(obj.Sigma(:,:,c))) < size(block,2))
                fprintf('%s: number of different variances in SIGMA differs from dimensionality\n',progname);
            end
        end
        model{i} = obj;
        if (debug)
            tim = toc;
            fprintf('%s: %0.2f [s]\n', progname, tim);
        end
    end
end




    
