function D = emmfl(data,varargin)
% EMMFL computes the feature likelihood as proposed in the paper
%
%   Vasconcelos, N., "A probabilistic architecture for Content-Based 
%   Image Retrieval", CVPR'00
%
%   D = EMMFL(DATA,MODELS,'PARAM1',val1, ...) computes the distance
%   matrix D where the (i,j) position denotes the distance between EMM
%   models (contained in MODELS) i and j, using the feature likelihood.
%
%   Parameters are:
%
%       DATA                Is a cell array with elements 'image'
%                           'dim', 'idx', 'filename'. 'image' hold the
%                           color or grayscale image. 'dim' specifies the
%                           dimension of the image (in case of color images
%                           'dim' is 3, otherwise 1). 'idx' is not used
%                           here; in case of image classification
%                           it can be used to specify the class the
%                           image belongs to. Finally, 'filename' holds the
%                           image filename
%
%       MODELS              Is a cell array containing the same number of models
%                           as there are images in DATA. The models are
%                           fitted multivariate Gaussian Mixture models
%                           fitted using MATLABs gmdistribution.fit routine
%
%       'preprocess'        true, if you want to preprocess the images
%                           first, subtracting mean and div. by std. dev.
%
%       'step'              keep the coefficients of every every 'step'-th
%                           block (default: 1 => 8x8 blocksize gives 256
%                           feature vectors)
%
%       'debug'             true, if you want to obtain debug messages
%                           (default: false)
%
%   Example:
%       
%        E = emmfl(data,models,'debug',true,'preprocess',true,'step',2);
%
%   Author: Roland Kwitt, rkwitt@gmx.at, 2009

    p = inputParser;
    p.addRequired('data',@iscell);
    p.addParamValue('preprocess',false,@islogical);
    p.addParamValue('grayscale',false,@islogical);
    p.addParamValue('step',1,@isnumeric);
    p.addRequired('models',@iscell);
    p.addParamValue('debug',false,@islogical);
    p.parse(data,varargin{:});
    models = p.Results.models;
    preprocess = p.Results.preprocess;
    debug = p.Results.debug;
    grayscale = p.Results.grayscale;
    step = p.Results.step;
    progname = 'emmfl';
    fun = @dct2;
    for i=1:length(data) % queries
        if (debug)
            fprintf('%s: processing query %d\n',progname, i);
        end
        im = data{i}.image;
        if (data{i}.dim > 1) % image is a color image
            if (grayscale)  % convert to grayscale
                if (debug)
                    fprintf('%s: converting to grayscale\n', progname);
                end
                im = double(rgb2gray(im));
                if (preprocess) % preprocess luminance image
                    if (debug)
                        fprintf('%s: subtracting mean and dividing by std. deviation\n',progname);
                        im = (im - mean2(im))./std2(im);
                    end
                end
            else % leave color image but convert to YBR
                if (debug)
                    fprintf('%s: converting RGB to YCbCr\n',progname);
                    im = double(rgb2ycbcr(im));
                end
                if (preprocess) % preprocess each channel
                    if (debug)
                        fprintf('%s: subtracting mean and dividing by std. deviation\n',progname);    
                    end
                    for c=1:size(im,3)
                        channel = im(:,:,c);
                        channel = (channel - mean2(channel))./std2(channel);
                        im(:,:,c) = channel;
                    end
                end
            end
        else % image is a grayscale image
            if (preprocess) % preprocess luminance component
                if (debug)
                    fprintf('%s: subtracting mean and dividing by std. deviation\n',progname);
                    im = (im - mean2(im))./std2(im);
                end
            end
        end
        
        % after preprocessing we extract DCT coefficients from
        % non-overlapping blocks of each channel. in case of grayscale
        % image we just have one channel of course;
        
        for c=1:size(im,3)
            J(:,:,c) = blkproc(im(:,:,c),[8 8],fun);
        end
        X = exzz(J,'blk',8,'coefficients',1:16,'debug',debug);
        
        % next, we reduce the number of feature vectors according to 'step'
        % which controls which non-overlapping blocks we consider;
        
        if (debug)
            fprintf('reducing feature %d feature vectors to %d\n', size(X,1), length(1:step:size(X,1)));
        end
        X = X(1:step:size(X,1),:);
        
        % finally, we calculate the feature likelihood of the extracted
        % feature vectors according to the candiate image models;
        
        for j=1:length(data) %candidates
            if (i==j)
                D(i,j) = 0;
                continue;
            end
            D(i,j) = sum(log(pdf(models{j},X)));
        end
    end
end
