function vec = exzz(A,varargin)
% EXZZ returns a matrix of extracted DCT coefficients from non-overlapping
% blocks of an input matrix.
%   VEC = EXZZ(A,'PARAM1',val1, ...) computes a matrix NxM matrix VEC. N
%   denotes the number of extracted 8x8 blocks from image A and M denotes the number of
%   extracted coefficients of each 8x8 block. Coefficients are extracted in
%   zigzag scan order.
%
%   Parameters are:
%
%   A               grayscale image; only grayscale supported yet
%
%   'blk'           blocksize in one dimension. Only square blocks are
%                   supported (default: 8).
%
%   'coefficients'  number of coefficients to extract in zigzag scan order.
%                   (default: 1:16)
%
%   'debug'         true, if you want to obtain debug messages
%                   (default: false)
%
%   Author: Roland Kwitt, rkwitt@gmx.at, 2009

    p = inputParser;
    p.addRequired('A',@(x)~isempty(x));
    p.addParamValue('blk',8,@(x)x>=1 && x<= 16);
    p.addParamValue('coefficients', 1:16, @isvector);
    p.addParamValue('debug',false,@islogical);
    p.parse(A,varargin{:});
    blk = p.Results.blk;
    coefficients = p.Results.coefficients;
    debug = p.Results.debug;
    progname = 'exzz';

    cnt = 1; 
    ind = zigzag(blk);
    for i=1:blk:size(A,1)
        for j=1:blk:size(A,2)
            tmp = []; B = [];
            % extract block
            B = A(i:i+blk-1,j:j+blk-1,:);
            for z = coefficients
                coefs = B(ind(z,1),ind(z,2),:);
                tmp = [tmp coefs(:)'];
            end
            vec(cnt,:) = tmp;
            cnt = cnt + 1;
        end
    end
    if (debug)
        fprintf('%s: extracted %d non-overlapping blocks\n',progname,cnt-1);
    end
end





