function D = klapprox(Q,I,C,N,method)
% KLAPPROX computes two approximations to the KL-divergence between
% Gaussian Mixture Models (GMMs)
%   D = KLAPPROX(Q,I,C,N,METHOD) computes the KL-divergence approximation
%   between the Gaussian Mixture Models Q and I (fitted by
%   gmdistribution.fit) with C components and dimensionality N. The
%   approximationa approach is either 'vasc' or 'gold' referring to the
%   approximations of Vasconcelos (Asymptotic Likelihood Approximation) in
%   the paper
%
%   Vasconcelos, N., "On the efficient evaluation of probabilistic similarity
%   functions for image retrieval", IEEE Transactions on Information
%   Theory, volume 50, issue 7, 2004
%
%   and the approximation given by Goldberger et al. in the paper
%
%   Goldberger, J. and Gordon, S. and Greenspan, H., "An Efficient
%   Similarity Measure Based on Approximations of the Kullback-Leibler
%   divergence between two Gaussian Mixtures", In Proceedings of the IEEE
%   International Conference on Computer Vision (ICCV'03), 2003
%
%   Author: Roland Kwitt (rkwitt at gmx dot at)


    D = 0;
    for j=1:C
        switch method
            case 'vasc'
                d(j) = vascalign(Q,I,C,j);
            case 'gold'
                d(j) = goldalign(Q,I,C,j,N);
        end
        D = D + Q.PComponents(j) * ...
            (kldivgauss(Q.mu(j,:),I.mu(d(j),:), ...
            diag(Q.Sigma(:,:,j)),diag(I.Sigma(:,:,d(j))),N) +  ...
            log(Q.PComponents(j)/I.PComponents(d(j))) ...
            );
    end
end

% Mahalanobis distance (Vasconcelos), see Gold03
function ind = vascalign(Q,I,C,i)
   x = Q.mu(i,:);
   for j=1:C
       y = I.mu(j,:);
       S = diag(I.Sigma(:,:,j));
       dist(j) = (x-y)*inv(S)*(x-y)';
   end
   [mini,ind] = min(dist);
end

% Better approximation proposed in Gold03
function ind = goldalign(Q,I,C,i,N)
    dist = [];
    for j=1:C
        dist(j) = kldivgauss(Q.mu(i,:),I.mu(j,:), ...
            diag(Q.Sigma(:,:,i)),diag(I.Sigma(:,:,j)),N) - log(I.PComponents(j));
    end
    [mini,ind] = min(dist);
end




