#include <stdlib.h>
#include <stdio.h>
#include <string.h>
#include <assert.h>
#include <sys/types.h>
#include <gsl/gsl_matrix.h>
#include <gsl/gsl_vector.h>
#include <gsl/gsl_math.h>
#include <gsl/gsl_linalg.h>
#include <gsl/gsl_complex.h>
#include <gsl/gsl_complex_math.h>
#include <gsl/gsl_blas.h>
#include <gsl/gsl_cblas.h>
#include <gsl/gsl_permutation.h>
#include <math.h>
#include "diststore.h"
#include "common.h"
#include "pthread.h"
#include "getopt.h"


// globals
const char *progname = "ala";
static unsigned int verbose = 0;

static double symdet(gsl_vector *);
static void klapprox(gsl_vector ***, gsl_vector ***, gsl_vector **, gsl_vector***, double **, unsigned int, unsigned int, unsigned int, unsigned int, unsigned int);
static double klgauss(gsl_vector *, gsl_vector *, gsl_vector *, gsl_vector *, gsl_vector *, gsl_vector *, double, double );
static int ala(gsl_vector **Q, gsl_vector **I, gsl_vector **muQ, gsl_vector **muI, gsl_vector *piQ, gsl_vector *piI, gsl_vector **invQ, gsl_vector **invI, double *logdetQ, double *logdetI, unsigned int C, unsigned int c);


static void *klapprox_thread(void *thread_params);
typedef struct {
    gsl_vector ***covlist;
    gsl_vector ***mulist;
    gsl_vector **pilist;    
    gsl_vector ***invlist;
    double **detlist;
    unsigned int C;
	unsigned int S;
} metric_params_t;

// default values
const char *DIST_OUT_FNAME = "dist.bin";
const char *LIST_FNAME = "filelist.txt";
const char *MODEL_FNAME = "";
const unsigned int DEFAULT_COMPONENT_NUM = 8; // eight mixture components
const unsigned int DEFAULT_COMPONENT_DIMENSIONALITY = 64; // i.e. 64 DCT coefficients (including DC)
const unsigned int DEFAULT_START_IMAGE = 0;
const unsigned int DEFAULT_PROCESS_IMAGES = 0;
const unsigned int DEFAULT_NUM_THREADS = 1;
const unsigned int DEFAULT_SYMMETRY = 0;

static void usage() {
    fprintf(stderr,
        "usage: %s [-C n] [-P n] [-d file] [-l file] [-s n] [-c n] [-S 0,1] [-t n] [-m mode]\n"
        "\n"
        "\t-C n\t\tnumber of mixture components\n\t\t\t(default: %d)\n"
        "\t-P n\t\tdimensionality of multivariate gaussian components\n\t\t\t(default: %d)\n"
        "\t-d file\t\tfile name of output distance matrix\n\t\t\t(default: %s)\n"
        "\t-l file\t\tfile name with list of files to process\n\t\t\t(default: %s)\n"
        "\t-s n\t\timage to start from, should be multiple of 16 (default: 0)\n"
        "\t-c n\t\tnumber of images to process (default: 0 .. all)\n"
		"\t-S 0,1\tsymmetric '0' (no) or '1' (yes) (default: %d)\n"
        "\t-t n\t\tnumber of threads to use (default: 1)\n"                        
        "\t-m mode\t\tcomputation mode 'matrix' or 'bestn' (default: 'matrix')\n"
        "\t-v\t\tverbose\n",
        progname, DEFAULT_COMPONENT_NUM, DEFAULT_COMPONENT_DIMENSIONALITY,DIST_OUT_FNAME, LIST_FNAME,DEFAULT_SYMMETRY);
}

int main(int argc, char *argv[]) {
    int c;
    unsigned int num, cnt;
    gsl_vector *mu = NULL; // mean vector
	gsl_vector *cov = NULL; // covariance matrix(diagonal, thus vector) 
	gsl_vector *pi = NULL; // weights
    gsl_vector **pilist = NULL; // holds one weight vector per image
	gsl_vector ***mulist = NULL; // holds a vector for each component and image
	gsl_vector ***covlist = NULL; // holds a covariance matrix for each component and image
	gsl_vector ***invlist = NULL; // holds inverse of covariance matrices for further use
	double **detlist = NULL; // holds log-determinants of covariance matrices for further use

	FILE *mfile = NULL; // general file descriptor

    unsigned int C = DEFAULT_COMPONENT_NUM;
    unsigned int P = DEFAULT_COMPONENT_DIMENSIONALITY; 
	unsigned int S = DEFAULT_SYMMETRY;
    unsigned int start_image = DEFAULT_START_IMAGE;
    unsigned int process_images = DEFAULT_PROCESS_IMAGES;
    unsigned int num_threads = DEFAULT_NUM_THREADS;
        
    char dist_out_fname[1024]; strcpy(dist_out_fname, DIST_OUT_FNAME);
    char list_fname[1024]; strcpy(list_fname, LIST_FNAME);
    char model_fname[1024]; strcpy(model_fname, MODEL_FNAME);    
    compmode_t mode = DISTMATRIX;

    while ((c = getopt(argc, argv, "C:P:d:l:L:s:S:c:m:t:vh")) != EOF) {
        switch (c) {
            case 'C':
                C = atoi(optarg);
                break;
            case 'P':
                P = atoi(optarg);
                break;
			case 'S':
				S = atoi(optarg);
				break;
            case 'd':
                strcpy(dist_out_fname, optarg);
                break;
            case 'l':
                strcpy(list_fname, optarg);
                break;
            case 'L':
                strcpy(model_fname, optarg);
                break;
            case 's':
                start_image = atoi(optarg);
                break;
            case 'c':
                process_images = atoi(optarg);
                break;
            case 'm':
                mode = check_compmode(optarg);
                break;
            case 't':
                num_threads = atoi(optarg);
                break;
            case 'h':
            case '?': 
                usage();
                exit(0);                                           
            case 'v':
                verbose = 1;
                break;
        }
    }
    argc -= optind;
    argv += optind;

    if (strlen(model_fname) == 0) {
        FILE *listfile = 0;
        if (!open_models_filelist(list_fname, &num, &listfile)) {
            exit(EXIT_FAILURE);
        }

	    // allocate enough space for all images
     	pilist = (gsl_vector **)malloc(num * sizeof(gsl_vector *));
     	covlist = (gsl_vector ***)malloc(num * sizeof(gsl_vector **));
     	invlist = (gsl_vector ***)malloc(num * sizeof(gsl_vector **));
     	mulist = (gsl_vector ***)malloc(num * sizeof(gsl_vector **));
	    detlist = (double **)malloc(num * sizeof(double *));
        
	    cnt = 0;
        if (verbose) fprintf(stderr, "%s: trying to read %d models ...\n", progname, num);
        char line[1024]; // whole line
	    while (fgets(line, sizeof(line), listfile) != NULL) {
        	char modelname[1024]; // name of model

            if (strlen(line) <= 1) continue;
	        line[strlen(line)-1] = '\0';

      		if (verbose)
                fprintf(stderr, "%s: reading parameters (pi,mu,cov) for %d components, model %s\n", progname, C, line);

		    // open weight file, read entries to vector, close file
		    sprintf(modelname, "%s.pi", line);
		    mfile = fopen(modelname, "rb");
		    if (!mfile) {
               	fprintf(stderr, "%s: failed to open '%s': %s\n", progname, modelname, strerror(errno));
               	exit(EXIT_FAILURE);
		    }
		
		    pi = gsl_vector_calloc(C);
		    if (gsl_vector_fread(mfile, pi) == GSL_EFAILED) {
                	fprintf(stderr, "%s: failed to read '%s': %s\n", progname, modelname, strerror(errno));
                	exit(EXIT_FAILURE);
		    }
		    pilist[cnt] = pi;
		    fclose(mfile);

		    // alloc space for C mean vectors and C covariance matrices
       	 	mulist[cnt] = (gsl_vector **)malloc(C * sizeof(gsl_vector *));
       	 	covlist[cnt] = (gsl_vector **)malloc(C * sizeof(gsl_vector *));
       	 	invlist[cnt] = (gsl_vector **)malloc(C * sizeof(gsl_vector *));
		    detlist[cnt] = (double *)malloc(C * sizeof(double));
          
		    for (unsigned int m = 0; m < C; m++) {
			    sprintf(modelname, "%s.%d.mu", line, m);
			    mfile = fopen(modelname, "rb");
        		if (!mfile) {
                   	fprintf(stderr, "%s: failed to open '%s': %s\n", progname, modelname, strerror(errno));
                   	exit(EXIT_FAILURE);
        		}
			    mu = gsl_vector_calloc(P);
			    if (gsl_vector_fread(mfile, mu) == GSL_EFAILED) {
                	fprintf(stderr, "%s: failed to read '%s': %s\n", progname, modelname, strerror(errno));
                	exit(EXIT_FAILURE);
			    }
			    fclose(mfile);
			    mulist[cnt][m] = mu;

			    sprintf(modelname, "%s.%d.cov", line, m);
			    mfile = fopen(modelname,"rb");
        		if (!mfile) {
                   	fprintf(stderr, "%s: failed to open '%s': %s\n", progname, modelname, strerror(errno));
                   	exit(EXIT_FAILURE);
        		}
			    cov = gsl_vector_calloc(P);
			    if (gsl_vector_fread(mfile, cov) == GSL_EFAILED) {
                	fprintf(stderr, "%s: failed to read '%s': %s\n", progname, modelname, strerror(errno));
                	exit(EXIT_FAILURE);
			    }
		        fclose(mfile);
		        
			    covlist[cnt][m] = cov;
               	// detlist[cnt][m] = log(symdet(cov)); // sum(abs(cov)) - abs doesnt matter since, variance is >= 0
				detlist[cnt][m] = symdet(cov); 
                gsl_vector *inverse = gsl_vector_calloc(P);
                gsl_vector_set_all (inverse, 1.0); // set all elements to 1.0
                gsl_vector_div(inverse,cov); // divide all 1.0 by elements of covariance mat. diagonal --> inverse
                //gsl_vector_fprintf(stderr, inverse, "%0.5f");
                invlist[cnt][m] = inverse;
            }
		    cnt++;
	    }
	    fclose(listfile);
    }
    else {
        FILE *mfile = 0;

        if (!open_models(model_fname, &num, (2*C*P+C)*sizeof(double), &mfile)) {
            exit(EXIT_FAILURE);
        }

	    // allocate enough space for all images
     	pilist = (gsl_vector **)malloc(num * sizeof(gsl_vector *));
     	covlist = (gsl_vector ***)malloc(num * sizeof(gsl_vector **));
     	invlist = (gsl_vector ***)malloc(num * sizeof(gsl_vector **));
     	mulist = (gsl_vector ***)malloc(num * sizeof(gsl_vector **));
	    detlist = (double **)malloc(num * sizeof(double *));

        if (verbose) fprintf(stderr, "%s: trying to read %d models ...\n", progname, num);

        for (cnt = 0; cnt < num; cnt++) {

      		if (verbose && (cnt % 1000) == 0)
                fprintf(stderr, "%s: reading data for model %d\n", progname, cnt);                

       	 	mulist[cnt] = (gsl_vector **)malloc(C * sizeof(gsl_vector *));
       	 	covlist[cnt] = (gsl_vector **)malloc(C * sizeof(gsl_vector *));
       	 	invlist[cnt] = (gsl_vector **)malloc(C * sizeof(gsl_vector *));
		    detlist[cnt] = (double *)malloc(C * sizeof(double));

		    for (unsigned int m = 0; m < C; m++) {
			    cov = gsl_vector_calloc(P);
			    if (gsl_vector_fread(mfile, cov) == GSL_EFAILED) {
                	fprintf(stderr, "%s: failed to read covariance model %d.%d: %s\n", progname, cnt, m, strerror(errno));
                	exit(EXIT_FAILURE);
			    }
		        
			    covlist[cnt][m] = cov;
               	//detlist[cnt][m] = log(symdet(cov)); // sum(abs(cov)) - abs doesnt matter since, variance is >= 0
               	detlist[cnt][m] = symdet(cov); // sum(abs(cov)) - abs doesnt matter since, variance is >= 0
                gsl_vector *inverse = gsl_vector_calloc(P);
                gsl_vector_set_all(inverse, 1.0); // set all elements to 1.0
                gsl_vector_div(inverse, cov); // divide all 1.0 by elements of covariance mat. diagonal --> inverse
                //gsl_vector_fprintf(stderr, inverse, "%0.5f");
                invlist[cnt][m] = inverse;

			    mu = gsl_vector_calloc(P);
			    if (gsl_vector_fread(mfile, mu) == GSL_EFAILED) {
                	fprintf(stderr, "%s: failed to read mu model %d.%d: %s\n", progname, cnt, m, strerror(errno));
                	exit(EXIT_FAILURE);
			    }
			    mulist[cnt][m] = mu;
            }

		    pi = gsl_vector_calloc(C);
		    if (gsl_vector_fread(mfile, pi) == GSL_EFAILED) {
                	fprintf(stderr, "%s: failed to read pi model %d: %s\n", progname, cnt, strerror(errno));
                	exit(EXIT_FAILURE);
		    }
		    pilist[cnt] = pi;
        }

        fclose(mfile);    
    }
    
    if (!init_diststore(num, (mode == DISTMATRIX) ? MATRIX : BOTTOMN)) {
        fprintf(stderr, "%s: failed to initialize distance store, exit.\n", progname);
        exit(EXIT_FAILURE);
    }

    metric_params_t metric_params;
    metric_params.covlist = covlist;
    metric_params.mulist = mulist;
    metric_params.pilist = pilist;            
    metric_params.invlist = invlist;        
    metric_params.detlist = detlist;        
    metric_params.C = C; // components 
	metric_params.S = S; // use symmetric
    
    if (!start_threads(num_threads, num, start_image, process_images, klapprox_thread, &metric_params)) {
        fprintf(stderr, "%s: failed to start thread(s), exit.\n", progname);
        exit(EXIT_FAILURE);
    }

    if (!write_diststore(dist_out_fname)) {
        fprintf(stderr, "%s: failed to write distance store, exit.\n", progname);
        exit(EXIT_FAILURE);
    }
    
    free_diststore();

	// free memory
    for (unsigned int i = 0; i < cnt; i++) {
		gsl_vector_free(pilist[i]);
		for (unsigned int m = 0; m < C; m++) {
    		gsl_vector_free(mulist[i][m]);
			gsl_vector_free(covlist[i][m]);
			gsl_vector_free(invlist[i][m]);
		}
		free(detlist[i]);
		free(mulist[i]);
		free(covlist[i]);
		free(invlist[i]);
    }
	free(mulist);
	free(detlist);
	free(invlist);
	free(covlist);
	free(pilist);
	
    return EXIT_SUCCESS;
}

static double symdet(gsl_vector *x) {
    //double prod = 1.0;
	double logsum = 0.0;
    for (unsigned int a = 0; a < x->size; a++)
		logsum += log(x->data[a]);
	return logsum;
        //prod *= x->data[a];
    //return prod;
}


// Kullback-Leibler divergence between two Gaussian distributions 
static double klgauss(gsl_vector *mua, gsl_vector *mub, gsl_vector *A, gsl_vector *B, gsl_vector *invA, gsl_vector *invB, double logdetA, double logdetB) 
{
    double d = 0.0;
    double trace = 0.0;
    double mahal = 0.0;
	gsl_vector *prodInvAB = gsl_vector_calloc(A->size);
    gsl_vector_memcpy(prodInvAB,invB); // copy inv(B) 
    gsl_vector_mul(prodInvAB,A); // inv(B)*A    
	trace = gsl_blas_dasum(prodInvAB);
    //fprintf(stderr, "trace: %0.5f\n", trace);
        
    gsl_vector *tmub = gsl_vector_calloc(mub->size);
	gsl_vector_memcpy(tmub,mub);
	gsl_vector_sub (tmub,mua); // tmub = mub - mua
	gsl_vector *term1 = gsl_vector_calloc(tmub->size);
    gsl_vector_memcpy(term1,invB);
    gsl_vector_mul(term1,tmub);
	gsl_blas_ddot(term1, tmub, &mahal); 
	d = 1.0/2.0 * (logdetB - logdetA + trace + mahal - (double)mua->size);
	
	// free the memory
	gsl_vector_free(prodInvAB);
	gsl_vector_free(tmub);
	gsl_vector_free(term1);
	return d;
}

// allignement used by vasconcelos (for the "asymptotic likelihood approximation")
static int ala(gsl_vector **Q, gsl_vector **I, gsl_vector **muQ, gsl_vector **muI, gsl_vector *piQ, gsl_vector *piI, gsl_vector **invQ, gsl_vector **invI, double *logdetQ, double *logdetI, unsigned int C, unsigned int c) {
	gsl_vector *tmuQ = gsl_vector_calloc(muQ[c]->size);
	gsl_vector *dist = gsl_vector_calloc(C);
	gsl_vector *prod = gsl_vector_calloc(tmuQ->size);
	double mahal = 0.0;
	for (unsigned int j = 0; j < C; j++) {
		gsl_vector_memcpy(tmuQ,muQ[c]);
		gsl_vector_sub(tmuQ,muI[j]); 
		gsl_vector_memcpy(prod,tmuQ);
		gsl_vector_mul(prod,invI[j]);
		gsl_blas_ddot(prod,tmuQ,&mahal);
		gsl_vector_set(dist,j,mahal);
	}		
	unsigned int i = gsl_vector_min_index(dist);
	gsl_vector_free(dist);
	gsl_vector_free(prod);
	gsl_vector_free(tmuQ);
	return i;
}

// approximation of the KL-divergence between multivariate Gaussian mixture models
static void klapprox(gsl_vector ***covlist, gsl_vector ***mulist, gsl_vector **pilist, 	gsl_vector ***invlist, double **detlist, unsigned int num, unsigned int C, unsigned int start_image, unsigned int process_images, unsigned int usesym) {
	for (unsigned i = start_image; i < start_image+process_images; i++) { // iterate over all images (queries)
    	if (verbose)  {
            fprintf(stderr, "%s: processing query %d, %.2f %% done\n", progname, i, (i-start_image)*100.0/process_images);
		}
		gsl_vector **Q = covlist[i]; // covariances for model i
		gsl_vector **muQ = mulist[i]; // mean vectors for model i
		gsl_vector **invListQ = invlist[i]; // inverse covariances for model i
		double *logdetQ = detlist[i]; // cov. matrix determinants (log) of model i
		int queryPIisempty = gsl_vector_isnull(pilist[i]);
		if (queryPIisempty) { // if queries' pis' are empty, so is the model
			fprintf(stderr, "Query %d is empty\n", i);
			for (unsigned int j=0; j <num; j++) {
				if (i == j) {
					store_dist(i,j,0.0);
					continue;
				}
				store_dist(i,j,DBL_MAX);
			}
			continue;
		}
		for (unsigned int j = 0; j < num; j++) { // iterate over all images (candidates)
			if (i == j) {
				store_dist(i,j,0.0);
				continue;
			}
			gsl_vector **I = covlist[j]; 
			gsl_vector **muI = mulist[j];	
			gsl_vector **invListI = invlist[j];
			double *logdetI = detlist[j];
			double d = 0; double d1 = 0.0; double d2 = 0.0;
			int candidatePIisnull = gsl_vector_isnull(pilist[j]);
			if (candidatePIisnull) { // candidates' pis are null, so is the candidate model
				store_dist(i,j,DBL_MAX);
				continue;
			}
			int selector;
			for (unsigned int c = 0; c < C; c++) { // iterate over all C components 
				if (usesym) {
					selector = ala(Q,I,muQ,muI,pilist[i],pilist[j],invListQ,invListI,logdetQ, logdetI, C, c);
					d1 = d1 + gsl_vector_get(pilist[i],c) * 
                    	(klgauss(muQ[c],muI[selector],Q[c],I[selector], invListQ[c],invListI[selector], logdetQ[c], logdetI[selector]) + 
            			log(gsl_vector_get(pilist[i],c)) - log(gsl_vector_get(pilist[j],selector)));	
				
					selector = ala(I,Q,muI,muQ,pilist[j],pilist[i],invListI,invListQ,logdetI, logdetQ, C, c);
					d2 = d2 + gsl_vector_get(pilist[j],c) * 
						(klgauss(muI[c],muQ[selector],I[c],Q[selector], invListI[c],invListQ[selector], logdetI[c], logdetQ[selector]) + 
						log(gsl_vector_get(pilist[j],c)) - log(gsl_vector_get(pilist[i],selector)));
					// average to obtain a symmetric similarity measure 
					d = 1.0/2.0 * (d1 + d2);
				} else {
					selector = ala(Q,I,muQ,muI,pilist[i],pilist[j],invListQ,invListI,logdetQ, logdetI, C, c);
					d = d + gsl_vector_get(pilist[i],c) * 
                    	(klgauss(muQ[c],muI[selector],Q[c],I[selector], invListQ[c],invListI[selector], logdetQ[c], logdetI[selector]) + 
            			log(gsl_vector_get(pilist[i],c)) - log(gsl_vector_get(pilist[j],selector)));	
				}
			}
			store_dist(i, j, d);
		}
	}
}

static void *klapprox_thread(void *thread_params) {
    if (!thread_params) 
        return 0;
        
    thread_params_t *p = (thread_params_t *) thread_params;
    metric_params_t *m = (metric_params_t *) p->metric_params;
    
    if (!m)
        return 0;
       
    if (verbose)
        fprintf(stderr, "%s: starting thread %d, processing images %d to %d (%d images)\n", progname, p->thread_num, p->start_image, p->start_image+p->process_images-1, p->process_images);

    klapprox(m->covlist, m->mulist, m->pilist, m->invlist, m->detlist, p->num_images, m->C, p->start_image, p->process_images, m->S);	

    return 0;
}

